#include "stdafx.h"

int main(int argc, char** argv) {
	
	google::InitGoogleLogging(argv[0]);
	
	std::string imageName, calibration = "chrome";
	int imageScale = 1, cropping = 0;
	bool showTexture = true;

	if (argc >= 2) { // The default calibration path will be used
		// Read the file name from the command line argument and convert it from a char array to a string
		const char *argument1 = argv[1];
		imageName = argument1;
	}
	if (argc >= 3) { // The specified calibration path will be used	
		const char *argument2 = argv[2];
		calibration = argument2;
	}
	if (argc >= 4) { // An image scale is provided	
		// Scale down the images for faster processing and rendering
		const char *argument3 = argv[3];
		std::stringstream stringValue;
		stringValue << argument3;
		stringValue >> imageScale;
		if (imageScale != 1)
			std::cout << "The images will be scaled to 1/" << imageScale << " of their original resolution in each axis." << std::endl;
	}
	if (argc == 5) { // An image cropping method is provided
		const char *argument4 = argv[4];
		std::stringstream stringValue2;
		stringValue2 << argument4;
		stringValue2 >> cropping;
		if (cropping == 1) {
			std::cout << "The images will be cropped to the model mask." << std::endl;
			showTexture = false;
		}
		else if (cropping == 2) {
			std::cout << "The images will be cropped to a user-defined region of interest." << std::endl;
			showTexture = false;
		}
		else {
			std::cout << "The images will not be cropped." << std::endl;
			cropping = 0;
		}
	}
	if ((argc < 2) || (argc > 5)) {
		std::cerr << "Enter the name of the object followed by the calibration set to use.\n\nThe images can be scaled down by providing a number.\nFor instance, entering 2 will scale the x and y axes by 1/2.\n\nThe model can be cropped to the mask, or to a specified region of interest.\nAs the last argument, enter the number corresponding to the desired option:\n0: No cropping. The full image area will be used.\n1: Automatically crop to the object mask.\n2: Use the mouse to draw a rectangle, then press the space bar to confirm." << std::endl;
		return -1;
	}

	std::cout << std::endl;


	const std::string calibrationPath = "/home/thomas/Documents/2017-12-04/" + calibration + "/" + calibration + ".";
	const std::string modelPath = "/home/thomas/Documents/2017-12-04/" + imageName + "/" + imageName + ".";
	const std::string texturePath = modelPath + "texture.png";


	// A vector of Mats allow multiple images to be allocated to the same object
	// textureImages are the full colour photographs that are used to create the texture
	// modelImages are the greyscale version of textureImages which are used to calculate the surface normals
	// calibrationImages contain the reflections of the chrome sphere, and are used to determine the light directions
	// charucoImages contain the ChArUco board, and are used to calibrate the camera
	std::vector< cv::Mat > textureImages, modelImages, calibrationImages, charucoImages;
	cv::Mat heightMap, normalMap, texture, lightDirections;
	cv::Rect calibrationBoundingBox;
	int width, height;

	// TODO: Reimplement cropping

	// Load the textureImages, modelImages, calibrationImages and charucoImages, then determine the bounding box around the chrome sphere
	loadImages(imageName, calibration, width, height, imageScale, textureImages, calibrationPath, calibrationImages, modelPath, modelImages, calibrationBoundingBox);

	// The number of lights is equal to the number of calibration images
	int numberOfLights = int(calibrationImages.size());
	// The rotation matrix is created from the rotation and translation vectors, and describes the rotation of the ChArUco board
	// The inverse of this transformation is applied to the light directions
	// The perspective transform warps the model images to align them with the ChArUco board
	std::vector< cv::Vec3d > rvecs, tvecs;
	cv::Mat rotationMatrix, rotationVector, charucoBoardImage, cameraMatrix, distortionCoefficients, perspectiveTransform, lightDirectionsPerspective = cv::Mat(numberOfLights, 3, CV_32F);

	cv::Ptr<cv::aruco::CharucoBoard> charucoBoard;
	
	charucoCalibration(width, height, charucoBoard, charucoBoardImage, cameraMatrix, distortionCoefficients);

	//charucoAlignment(charucoImages, textureImages, numberOfLights, width, height, rotationMatrix, rotationVector, perspectiveTransform, lightDirections, lightDirectionsPerspective, rvecs, tvecs, charucoBoard, charucoBoardImage, cameraMatrix, distortionCoefficients);

/*
	rotationMatrix = 
	rotationVector = 
	perspectiveTransform = 
	lightDirectionsPerspective = 
	std::vector< cv::Vec3d > & rvecs
	std::vector< cv::Vec3d > & tvecs
*/

	photometricStereo(imageName, calibration, modelPath, imageScale, cropping, showTexture, heightMap, normalMap, texture, lightDirections, lightDirectionsPerspective, textureImages, modelImages, calibrationImages, calibrationBoundingBox, width, height, rvecs, tvecs);
	//photometricStereo(imageName, calibration, modelPath, imageScale, cropping, showTexture, heightMap, normalMap, texture, lightDirections, lightDirectionsPerspective, charucoImages, modelImages, calibrationImages, calibrationBoundingBox, width, height, rvecs, tvecs);

	

	glm::vec3 position, lightInvDir;
	glm::mat4 depthProjectionMatrix, depthViewMatrix;
	float horizontalAngle, verticalAngle, FoV, SpecularIntensity = 0.5f, SpecularPower = 5.0f;
	GLuint programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularPowerID;
	std::vector<unsigned int> indices;
	int lightNumber = 0;
	double residual;

	initialiseOpenGL(heightMap, normalMap, texture, textureImages, lightDirections, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);
	//initialiseOpenGL(heightMap, normalMap, texture, textureImages, lightDirectionsPerspective, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);
	//initialiseOpenGL(heightMap, normalMap, texture, charucoImages, lightDirectionsPerspective, width, height, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower);

	/*
	glm::mat4 ModelMatrix = glm::mat4(1.0);
	glm::mat4 MVP, ViewMatrix, depthBiasMVP, ProjectionMatrix = glm::ortho(-0.5f * float(windowWidth) / float(windowHeight), 0.5f * float(windowWidth) / float(windowHeight), -0.5f, 0.5f, 0.0f, 10.0f);
	bool perspectiveProjection, shadowControl;
	computeResiduals(windowWidth, windowHeight, position, horizontalAngle, verticalAngle, FoV, ProjectionMatrix, ViewMatrix, lightInvDir, depthProjectionMatrix, depthViewMatrix, perspectiveProjection = false, lightDirections, textureImages, lightNumber, numberOfLights, SpecularIntensity, SpecularPower);
	*/
	
	
	
	do {
		openGL(lightNumber, depthProjectionMatrix, depthViewMatrix, width, height, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirections, textureImages, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, residual);
		//openGL(lightNumber, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirectionsPerspective, textureImages, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, residual);
		//openGL(lightNumber, depthProjectionMatrix, depthViewMatrix, position, horizontalAngle, verticalAngle, FoV, lightInvDir, lightDirectionsPerspective, charucoImages, programID, MatrixID, ModelMatrixID, ViewMatrixID, DepthBiasID, lightInvDirID, Texture, TextureID, depthTexture, ShadowMapID, vertexbuffer, uvbuffer, normalbuffer, elementbuffer, indices, depthProgramID, quad_programID, FramebufferName, quad_vertexbuffer, VertexArrayID, numberOfLights, SpecularIntensityID, SpecularIntensity, SpecularPowerID, SpecularPower, residual);
		
		// TODO: Check if the viewpoint has been changed
		
		specularMinimisation(SpecularIntensity, SpecularPower, residual);
	} while (glfwGetKey(window, GLFW_KEY_ESCAPE) != GLFW_PRESS &&	glfwWindowShouldClose(window) == 0); // Check if the ESC key was pressed or the window was closed
	
	
	
	
	
	terminateOpenGL(vertexbuffer, uvbuffer, normalbuffer, elementbuffer, programID, depthProgramID, quad_programID, Texture, FramebufferName, depthTexture, quad_vertexbuffer, VertexArrayID);
	
	
	return 0;
}
